/* Chapter 6.
	asc2unMM.c function: Memory Mapped implementation of the
	ASCII to UNICODE file copy function. */

#include "EvryThng.h"

BOOL Asc2Un (LPCTSTR fIn, LPCTSTR fOut, BOOL bFailIfExists)

/* ASCII to Unicode file copy function.
	fIn:		Source file pathname.
	fOut:		Destination file pathname.
	bFailIfExists:	Do not copy if the destination file already exists.
		Behavior is modeled after CopyFile. */
{
BOOL Complete = FALSE;
_try {
	HANDLE hIn, hOut;
	HANDLE hInMap, hOutMap;
	LPSTR pIn, pInFile;
	LPWSTR pOut, pOutFile;
	DWORD FsLow, fdwOut;

	/* Open the input file. */

	hIn = CreateFile (fIn, /*??*/, 0, NULL,
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hIn == INVALID_HANDLE_VALUE) 
		ReportException (_T ("\nFailure opening input file."), 1);

	/* Create a file mapping object on the input file. Use the file size. */

	hInMap = CreateFileMapping (/*??*/, NULL, /*??*/, 0, 0, NULL);
	if (hInMap == /*??*/)
		ReportException (_T ("Failure Creating input map."), 2);

	pInFile = MapViewOfFile (hInMap, /*??*/, /*??*/);
	if (pInFile == NULL)
		ReportException (_T ("Failure Mapping input file."), 3);

	/* Get the input file size. As the mapping succeeded, the file size is < 2 GB. */

	FsLow = GetFileSize (/*??*/L);
	if (FsLow == /*??*/)
		ReportException (_T ("Failure getting file size."), 4);

	/*  Create/Open the output file. */

	fdwOut = bFailIfExists ? CREATE_NEW : CREATE_ALWAYS;

	/* The output file MUST have Read/Write access for the mapping to succeed. */

	hOut = CreateFile (fOut, /*??*/E,
			0, NULL, fdwOut, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hOut == INVALID_HANDLE_VALUE) {
		Complete = TRUE; /* Do not delete an existing file. */
		ReportException (_T ("Failure Opening output file."), 5);
	}

	/* Set the output file's size to twice the input size. */

/*	if ( SetFilePointer (hOut, 2 * FsLow, NULL, FILE_BEGIN) == 0xFFFFFFFF
			|| !SetEndOfFile (hOut))
		ReportException (_T ("Failure setting file size.\n"), 6);
*/
	/* Map the output file. Using the file length in the next two calls
		is optional but must be done if you do not set the file size above. */
Question: What happens if the file size is less than the mapping size?

	hOutMap = CreateFileMapping (hOut, NULL, /*??*/, /*??*/, /*??*/, NULL);
	if (hOutMap == INVALID_HANDLE_VALUE)
		ReportException (_T ("Failure creating output map."), 7);
	pOutFile = MapViewOfFile (hOutMap, /*??*/, /*??*/, /*??*/);
	if (pOutFile == NULL)
		ReportException (_T ("Failure mapping output file."), 8);

	/* Now move the input file to the output file, doing all the work in memory. */

	pIn = pInFile;
	pOut = pOutFile;
	while (pIn < /*??*/) {
		*pOut = (WCHAR) *pIn;
		pIn++;
		pOut++;
	}

	/* Close all views and handles. */

	Unmap/*??*/;
	/*??*/;
	Complete = TRUE;
	return TRUE;
}

_except (EXCEPTION_EXECUTE_HANDLER) {
		/* Delete the output file if the operation did not complete successfully. */
	if (!Complete)
		DeleteFile (/*??*/);
	return FALSE;
}
}





