/*	MultiSemx.c  DEFECTIVE VERSION

	Copyright (c) 1998, Johnson M. Hart
	Permission is granted for any and all use providing that this copyright is
	properly acknowledged.
	There are no assurances of suitability for any use whatsosever.

	Library of synchronization "objects" to supplement the standard Win32
	events and mutexs. 
	
	For simplicity, this implementation shares a header file (Synchronize.h) with
	another library of synchronization objects (EvMxSem.c) that is designed
	to operate under Windows CE, which does not have semaphores.

	Implementation notes: 
	1.	All required internal data structures are allocated on the process's heap
	2.	Where appropriate, a new error code is returned (see the header
		file), or, if the error is a Win32 error, that code is unchanged.
	3.	Notice the new handle type, "SYNCHHANDLE" which has handles, counters,
		and	other information. This structure will grow as new objects are added
		to this set; some members are specific to only one or two of the objects;
		in particular, the structure is more general than is required just for
		some of the specific examples.
	4.	Mutexs are used for critical sections. These could be replaced with
		CRITICAL_SECTION objects if the objects are used totally within
		a process. Even better, if the object is not named, then you 
		know that it will be used only within the process and you could
		make the decision at create time. HOWEVER, you will loose the timeout
		option if you do so, AND you will also loose the naming and interprocess
		capabilities. Also, you will loose the opportunity to use
		SignalObjectAndWait.
	5.	These simulated semaphores can be named and hence can be shared
		between processes.
	6.	No Win32 semaphores are used, so this implementation (with null names)
		will operate under Windows CE.
	7.	The implementation shows serveral interesting aspect of synchronization, some
		of which are specific to Win32 and some of which are general. There are pointed
		out in the comments as appropriate.
	8.	These objects have a WaitForSingleObject equivalent. There is, however, no
		equivalent to WaitForMultipleObjects as this is very difficult, if not impossible
		to do efficiently outside of the kernel.
*/

#include "Everything.h"
#include "Synchronize.h"

static SYNCHHANDLE CleanUp (SYNCHHANDLE);
static SYNCHHANDLE AllocSynchHandle (LPCTSTR, LPTSTR, LPTSTR, LPTSTR, LPBOOL);
/*********************************************
	MULTIPLE WAIT SEMAPHORE function family.
	These semaphores allow a thread to wait atomically for several "units"
	rather than just one. The DEFAULT wait semantics are considered to be 
	"First Satisfiable". That is, a thread with a satisfiable request
	will be released even if a thread (regardless of priority) has an
	outstanding request that is not currently satisfiable.

	Optional semantics, specified at create time, are
	"First Come, First Served". A thread with request larger than the
	number of units currently available will block all subsequent
	threads, even those with satisfiable requests or those of
	higher priority. *** Be very careful as a large request may block
	(starve) all other threads. Do not wait for a large count, especially
	in FCFS mode, if you can not assure that the count will be satisifed
	eventually.
*/

SYNCHHANDLE CreateSemaphoreMW (

	LPSECURITY_ATTRIBUTES  lpSemaphoreAttributes,	/* pointer to security attributes */
    LONG  lInitialCount,	/* initial count */
    LONG  lMaximumCount,	/* maximum count */
	BOOL  fFirstComeFirstServed,	/*	First Satisfiable is the default */
    LPCTSTR  lpName )

/*	Multiple wait semaphore creation. 
	Requires a counter, a mutex to protect the semaphore state, and an
	autoreset event.

	Here are the rules that must always hold between the autoreset event
	and the mutex (any violation of these rules by the multiple wait semaphore
	functions will, in all liklihood, result in a defect):
		1.	No thread can set, pulse, or reset the event,
			nor can it access any part of the SYNCHHANDLE structure,
			without first gaining ownership of the mutex.
			BUT, a thread can wait on the event without owning the mutex
			(this is clearly necessary or else the event could never be set).
			This is a consequence of using a "signal" condition variable model.
		2.	The event is in a signaled state if and only if the count is
			greater than zero. To assure this property, the count should
			be checked after every semaphore decrease.
		3.	The semaphore count is always >= 0 and <= the maximum count
*/
{
}


BOOL ReleaseSemaphoreMW (SYNCHHANDLE hSemMW, LONG cReleaseCount, LPLONG lpPreviousCount)
/*	Multiple wait equivalent to ReleaseSemaphore */
{
}

DWORD WaitForSemaphoreMW (SYNCHHANDLE hSemMW, LONG cSemRequest, DWORD dwMilliseconds)
/*	Multiple wait semaphore equivalent of WaitForSingleObject.
	The second parameter is the number of units requested, and the waiting will be
	either first come, first served or first available, depending on the option
	selected at create time. */
{

}


BOOL CloseSynchHandle (SYNCHHANDLE hSynch)
/*	Close a synchronization handle. 
	Improvement: Test for a valid handle before closing the handle */
{
}


static SYNCHHANDLE CleanUp (SYNCHHANDLE hSynch)
{	/*	Prepare to return from a create of a synchronization handle.
		If there was any failure, free any allocated resouces
		"Flags" indicates which Win32 objects are required in the 
		synchronization handle */
}


static SYNCHHANDLE AllocSynchHandle (LPCTSTR lpName, 
	LPTSTR MutexName, LPTSTR EventName, LPTSTR MutexaName,
	LPBOOL pfNewObject)
/*	Allocate memory for a synchronization handle. Unnamed objects
	have their handles created directly in the process heap, whereas
	named objects have two handles, one allocated locally to
	contain the handles (which are local to this process) and
	out of a shared memory pool mapped to the paging file for
	the shared object state.
	Also, create the names for the three internal objects
	and determine if this is a new object that should be initialized. */
{
}

