/* Chapter 9. Client/Server. CLIENT VERSION. 
		Windows Sockets version.  */

/* Similar to timep, except the command is sent to a server process over a named pipe.
	Execute a command line (on the server) and display the response. */

/* This program illustrates:
	1. Windows sockets from the client side.
	2. Short-lived connections with a single request but multiple responses.
	3. Reading responses in server messages until the server disconnects. */

#define _NOEXCLUSIONS
#include "EvryThng.h"
#include "ClntSrvr.h"	/* Defines the request and response records. */

#define SERVER_PORT 1070   /* Well known port, also used by client. */

static void ReportWSError (LPCTSTR, DWORD);

struct sockaddr_in ClientSAddr;		/* Clients's Socket address structure */
WSADATA WSStartData;				/* Socket library data structure   */

int _tmain (int argc, LPTSTR argv [])
{
	SOCKET ClientSock = INVALID_SOCKET;

	LPTSTR targv = GetCommandLine ();
	REQUEST Request;	/* See ClntSrvr.h */
	RESPONSE Response;	/* See ClntSrvr.h */
	BOOL exit = FALSE, Disconnect = FALSE;
	int SockRev, ConVal, nRemainSend,nRemainRecv, nXfer;
	BYTE *pBuffer;

	if (targv == NULL)
		ReportError (_T ("Cannot read command line."), 1, TRUE);
	targv = SkipArg (targv);

		/* Put the command line in a request to the server. */

	_stprintf (Request.Record, _T("%s"), targv);

	/*	Initialize the WS library */
	SockRev = WSAStartup (/*??*/, &WSStartData);

		/*	Follow the standard client socket/connect sequence */
	ClientSock = socket(/*??*/, /*??*/, 0);
	if (ClientSock == /*??*/)
		ReportWSError (_T("Failed client socket() call"), 1);
    
    ClientSAddr.sin_family = /*??*/;	
    ClientSAddr.sin_addr./*??*/ = inet_addr( "10.1.12.192" ); /* Fill in properly */
    ClientSAddr.sin_port = /*??*/( SERVER_PORT );

	ConVal = connect (ClientSock, (/*??*/ *)&ClientSAddr, 
		sizeof(/*??*/));
	if (ConVal != 0) ReportWSError (_T("Failed client connect() call"), 2);

	/*	Write the request to the server. First, fill in the header. 
		The request is sent in two parts. First the header, then
		the command string proper. */
	nRemainSend = RQ_HEADER_LEN; 
	Request.Command = 1;
	Request.RqLen = (_tcslen (Request.Record) + 1) * TSIZE;
	pBuffer = (LPBYTE)&Request;
	/*	It is rare (never observed?) that the send is not completed
		on the first attempt, but, you can't take a chance. 
		It is also necessary to send the header and message
		body separately, conforming to the way it is received.
		In general, you would want to create a "message oriented"
		subroutine to do this work. */
	while (nRemainSend > 0)  {
		nXfer = send (/*??*/, pBuffer, nRemainSend, 0);
		if (nXfer == SOCKET_ERROR) ReportWSError (_T("client send() failed"), 3);
		if (nXfer == 0) {
			_tprintf (_T("\nServer disconnected before complete request header sent"));
			break;
		}
		nRemainSend -=nXfer; pBuffer += nXfer;
	}

	nRemainSend = Request.RqLen; 
	pBuffer = (LPBYTE)&Request.Record;
	while (nRemainSend > 0)  {
		nXfer = send (/*??*/, pBuffer, /*??*/, 0);
		if (nXfer == /*??*/) ReportWSError (_T("client send() failed"), 3);
		if (nXfer == /*??*/) {
			_tprintf (_T("\nServer disconnected before complete request sent"));
			break;
		}
		nRemainSend -=nXfer; pBuffer += nXfer;
	}


	/*  Request transmission complete */
	/*  Read the response records - there may be more than one.
		As each is received, write it to std out. */
	
	Disconnect = FALSE;	
	
	/*	if nRemainSen not 0, then the send failed
		nReaminSend never changes in the loop below, which is exited
		by disconnection */
	while (nRemainSend == 0 && !Disconnect)  { 
		/* Exit the loop when the server disconnects */
		/*	Read each response and send it to std out. 
			We have to read, first, the record header, and then
			read the rest of the record. You can't take anything for
			granted here.
			Comment: Named pipes are message oriented and make this
				whole problem much easier as you don't need to worry
				about detecting the message end. */
		
		/*  Read the header */
		nRemainRecv = RS_HEADER_LEN; pBuffer = (LPBYTE)&Response;
		while (nRemainRecv > 0 && !Disconnect) {
			/*  Note that the order of the while expression is important. Why?  */
			nXfer = recv (ClientSock, /*??*/);
			if (nXfer == SOCKET_ERROR) ReportWSError (_T("client header recv() failed"), 3);
			if (nXfer == 0) {
				Disconnect = TRUE;
			}
			else {
				nRemainRecv -=nXfer; pBuffer += nXfer;
			}
		}
		/*	Read the response record */
		nRemainRecv = Response.RsLen; 
		pBuffer = (LPBYTE)&Response.Record;
		while (nRemainRecv > 0 && !Disconnect) {
			nXfer = recv (ClientSock, pBuffer, nRemainRecv, 0);
			if (nXfer == SOCKET_ERROR) ReportWSError (_T("client response recv() failed"), 3);
			if (nXfer == 0) {
				Disconnect = TRUE;
			}
			else {
				nRemainRecv -=nXfer; pBuffer += nXfer;
			}
		}
		
		if (!Disconnect) 
			_tprintf (_T ("%s"), Response.Record);

	 }  /*  End of the Response read loop. Terminated only when
			the server disconnects */
	

	closesocket (/*??*/);
	_tprintf (_T("\n****End of server response. Leaving client"));
	WSA/*??*/;
	return 0;
}

static void ReportWSError (LPCTSTR Message, DWORD ExitCode)
/*	Similar to ReportError but with no formated message */
{
	DWORD ErrorCode = WSAGetLastError();

	_tprintf (_T("\n%s. Windows Sockets error code = %d"), Message, ErrorCode);
	if (ExitCode == 0) return;
	else ExitProcess (ExitCode);
	return;  /* Avoids a warning message */
}

