/* Chapter 9. Client/Server. SERVER PROGRAM. X Version
	Windows Sockets version.
	Execute the command in the request and return a response. 
		Server
	This is developed as a main program but could easily be turned into
	a Windows NT Service.
	Also, it is a simple single threaded implementation, capable of serving
	just one client at a time.
	Final note: This assumes TCP/IP as the underlying networking protocol,
	but it is easy to generalize to support any protocol.
	*/
#define _NOEXCLUSIONS
#include "evrythng.h"

/* ClntSrvr.h defines the message structures */
#include "ClntSrvr.h"

#define SERVER_PORT 1070   /* Well known port, also used by client. */

static void ReportWSError (LPCTSTR, DWORD);

struct sockaddr_in SrvSAddr;		/* Server's Socket address structure */
struct sockaddr_in ConnectSAddr;	/* Connected socket with client details   */
WSADATA WSStartData;				/* Socket library data structure   */


int _tmain (int argc, LPTSTR argv [])
{
	BOOL Done = FALSE, Disconnect = FALSE;
	/* Server listening and connected sockets. */
	SOCKET SrvSock = INVALID_SOCKET, ConnectSock = INVALID_SOCKET;
	HANDLE hTempFile;			/* Hold intermediate results */
	LONG nXfer, nRemainSend, nRemainRecv, AddrLen, HeaderSize = RQ_HEADER_LEN;
	STARTUPINFO StartInfo, StartInfoCh;
	PROCESS_INFORMATION ProcInfo;
	LPCTSTR TempFile = _T ("Results.tmp"); /* File to hold intermed results */
	FILE *fp = NULL;
	REQUEST Request;	/* Defined in ClntSrvr.h */
	RESPONSE Response;	/* Defined in ClntSrvr.h.*/
	int SockRev;		/* Windows Sockets revision */
	BYTE *pBuffer;
	SECURITY_ATTRIBUTES SaInheritable = {sizeof(SECURITY_ATTRIBUTES), NULL, TRUE }; 

	/* Startup info for the child and parent processes. */

	GetStartupInfo (&StartInfo);
	GetStartupInfo (&StartInfoCh);

	/*	Initialize the WS library */
	SockRev = WSAStartup (MAKEWORD (1, 1), &WSStartData);

	/*	Follow the standard server socket/bind/listen/accept sequence */
	SrvSock = socket(AF_INET, SOCK_STREAM, 0);
	if (SrvSock == INVALID_SOCKET) ReportWSError (_T("Failed server socket() call"), 1);
    
	/*	Prepare the socket address structure for binding the
    	server socket to the port number "reserved" for this service.
    	Accept requests from any client machine.  */

    SrvSAddr.sin_family = AF_INET;	
    SrvSAddr.sin_addr.s_addr = htonl( INADDR_ANY );    
    SrvSAddr.sin_port = htons( SERVER_PORT );	
	if (bind (SrvSock, (struct sockaddr *)&SrvSAddr, sizeof SrvSAddr) == SOCKET_ERROR)
		ReportWSError (_T("Failed server bind() call"), 1);
	if (listen (SrvSock, 5) != 0) ReportWSError (_T("Server listen() error"), 2);

	while (!Done) _try {		/* Main Command Loop with client connection. */
		_tprintf (_T ("\nServer is awaiting next request.\n"));
		
		AddrLen = sizeof(ConnectSAddr);
		ConnectSock = accept (SrvSock, (struct sockaddr *)&ConnectSAddr, 
			&AddrLen);
		if (ConnectSock == INVALID_SOCKET) {
			_tprintf (_T("\nServer connect() failed. %d. Try again"),
				WSAGetLastError());
			continue;  /*  Go back to attempt another connection. */
		}

		/*	Read the request. First the header, then the request text. */
		nRemainRecv = RQ_HEADER_LEN; 
		pBuffer = (LPBYTE)&Request;
		Disconnect = FALSE;
		while (nRemainRecv > 0 && !Disconnect) {
			/*  Note that the order of the while expression is important. Why?  */
			nXfer = recv (ConnectSock, pBuffer, nRemainRecv, 0);
			if (nXfer == SOCKET_ERROR) ReportWSError (_T("server header recv() failed"), 3);
			if (nXfer == 0) {
				Disconnect = TRUE;
			}
			else {
				nRemainRecv -=nXfer; pBuffer += nXfer;
			}
		}
		
		/*	Read the request record */
		nRemainRecv = Request.RqLen;

		pBuffer = (LPBYTE)&Request.Record;
		while (nRemainRecv > 0 && !Disconnect) {
			nXfer = recv (ConnectSock, pBuffer, nRemainRecv, 0);
			if (nXfer == SOCKET_ERROR) ReportWSError (_T("server request recv() failed"), 3);
			if (nXfer == 0) {
				Disconnect = TRUE;
			}
			else {
				nRemainRecv -=nXfer; pBuffer += nXfer;
			}
		}
		_tprintf (_T ("Request is: %s\n"), Request.Record);

		Done = (_tcscmp (Request.Record, _T ("$Quit")) == 0) || Disconnect;

		if (Done) continue;	/* Stop the server on "$Quit" command. */

		/* Create a process to execute the command.
			Its output will go to temporary file. */

		hTempFile = CreateFile (TempFile, GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				&SaInheritable, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

		StartInfoCh.hStdOutput = hTempFile;
		StartInfoCh.hStdError = hTempFile;
		StartInfoCh.dwFlags = STARTF_USESTDHANDLES;
		if (!CreateProcess (NULL, Request.Record, NULL, NULL,
				TRUE, 0, NULL, NULL, &StartInfoCh, &ProcInfo))
			PrintMsg (hTempFile,
					_T ("Server could not process request.\n"));
		else {
			CloseHandle (ProcInfo.hThread);
			WaitForSingleObject (ProcInfo.hProcess, INFINITE);
			CloseHandle (ProcInfo.hProcess);
		}
		CloseHandle (hTempFile);

		/* Send the temp file, one line at a time, with header, to the client. */

		fp = _tfopen (TempFile, _T ("r"));
		while ((_fgetts (Response.Record, MAX_RQRS_LEN, fp) != NULL)) {
			/*	Send the response up to the string end. Send in 
				two parts - header, then the response string. */
			nRemainSend = RS_HEADER_LEN; 
			Response.RsLen = (_tcslen (Response.Record) + 1) * TSIZE;
			pBuffer = (LPBYTE)&Response;
			while (nRemainSend > 0) {
				nXfer = send (ConnectSock, pBuffer, nRemainSend, 0);
				if (nXfer == SOCKET_ERROR) ReportWSError (_T("server send() failed"), 3);
				if (nXfer == 0) {
					_tprintf (_T("\nClient disconnected"));
					_leave;
				}
				nRemainSend -=nXfer; pBuffer += nXfer;
			}

			nRemainSend = Response.RsLen;
			pBuffer = (LPBYTE)&Response.Record;
			while (nRemainSend > 0) {
				nXfer = send (ConnectSock, pBuffer, nRemainSend, 0);
				if (nXfer == SOCKET_ERROR) ReportWSError (_T("server send() failed"), 3);
				if (nXfer == 0) {
					_tprintf (_T("\nClient disconnected"));
					_leave;
				}
				nRemainSend -=nXfer; pBuffer += nXfer;
			}

		}
	}

	/*  Request/Response complete. Prepare for next connection */
	_finally {
		if (fp != NULL) { fclose (fp); fp = NULL; }
		if (ConnectSock != INVALID_SOCKET) {
			closesocket (ConnectSock);
			ConnectSock = INVALID_SOCKET;
		}
		if (!DeleteFile (TempFile))
			ReportError (_T ("Failed to delete temp file."), 0, TRUE);
	}	/* End of while loop and server operation. Await another connection */

	closesocket (SrvSock);  /* Close the server listening socket */
	WSACleanup();

	return 0;
}

static void ReportWSError (LPCTSTR Message, DWORD ExitCode)
/*	Similar to ReportError but with no formated message */
{
	DWORD ErrorCode = WSAGetLastError();

	_tprintf (_T("\n%s. Windows Sockets error code = %d"), Message, ErrorCode);
	if (ExitCode == 0) return;
	else ExitProcess (ExitCode);
	return;  /* Avoids a warning message */
}

