/* Chapter 14. geonames_xxx (blank slate). July 13, 2010. Johnson M. Hart (Copyright 2010 JMH Associates, Inc.)
	Background: 
	1.  "LINQ to Objects Using C# 4.0" by Troy Magennis, Addison-Wesley, 2010. "LINQ Data Parallelism", pp 272ff.
	    This section describes a PLINQ solution to process an 825MB "geographical names" text file.
		This program will solve the same problem using native code and techniques from the next reference.
	2.  "Windows Parallelism, Fast File Searching, and Speculative Processing" by Johnson M. Hart, June 30, 2010.
	    http://www.informit.com/articles/article.aspx?p=1606242.
		The project Projects2008_64\compMP (source file CHAPTR14\compMP.c) implements the the solution discussed in the article.
	3.	"Sequential File Processing: A Fast Path from Native Code to .NET with Visits to Parallelism and Memory Mapped Files" by Johnson M. Hart, June 13, 2010.
		is a related C#/.NET solution (not using LINQ). The article will appear in Dr. Dobb's Journal in the future. A current copy (article, code, etc.)
		is at http://www.jmhartsoftware.com/SequentialFileProcessingSupport.html.
	4.  Windows System Programming Edition 4, Chapters 5, 7-10, and 14, describe the Windows API features that are used in this program.

	The code here uses native code to solve the same problem that Magennis (#1) solves with PLINQ so as to compare performance.
	Comment: Assume that we want to optimize performance, even if the code is less elegant and more difficult to write.

	Performance Results Overview, elapsed time (seconds) running on a 4-CPU system.

		#CPUs		Native, 64-bit code.		Native, 32-bit code.		PLINQ (#1 above), 32-bit code	C# 4.0/.NET 4.0
			1			8.613						7.678						28.569							3.142
			2			3.570						3.398						14.434							1.538
			4			1.937						1.845						10.630							0.849

	Quick conclusions:
		1.  The native code solution with memory mapped files is considrably faster than the PLINQ solution, but a factor of ~4
		2.  The PLINQ solution does not scale with the number of processors beyond 2 processors. The native code solution does scale well.
		3.  The native code, however, is far more complex and much less elegant than the PLINQ code. This complexity, in turn, can lead
			to maintenance difficulties and unpleasant bugs. The performance, however, may be well worth the trouble.
		4.  32-bit code is marginally faster than 64-bit code, but larger files than used in this example will require
			64-bit code.
		5.  A faster PLINQ solution may be possible, however, with a Memory Mapped File class that supports the IEnumberable interface.
			That will be a future project.
		6.  For the time being, it's safe to say that native code using threads and memory mapped files can provide excellent performance
			for numerous file processing problems. Furthermore, the basic coding pattern is well established with the WSP4 examples 
			and Background articles #2 and #3. Nonetheless, the code can become complex.
		7.  You can improve native code performance by setting the highest optimization level. You should be able to match the
		    C# 4.0/.NET performance.

/*  Problem overview: Find all locations with an elevation (column 15) > 8000 meters. Retain the name (column 1), the country (column 8)
    and the elevation  Sort in descenting order. Display the results (there are 16 places above 8000 meters. Only two names were familiar to me).
	You might also consider generalizing the query.
*/
/* geonames file [nThreads ] 
           file: pathname for the geonames database text file.
				Download and unzip from: http://download.geonames.org/export/dump/allCountries.zip. The unzipped file, allCountries.txt, is about 825MB.
				The fields are tab-separated, with one geographical location per line.
           nThreads: Number of independent threads (parallel work items to process the file)
                     Default: Number of processors on this system.
					 POTENTIAL ENHANCEMENT: If the file size is too large compared to physical memory, use the conventional]
						file processing solution. See "Windows Parallelism, Fast File Searching, and Speculative Processing"
						for more information.
					 ADDITIONAL ENHANCEMENT. Try thread pooling
 */
/* Example Output:
Mount Everest (8848m) - located in NP
K2 (8611m) - located in PK
Kanchenjunga (8586m) - located in NP
Lo-tzu Feng (8516m) - located in NP
Makalu (8463m) - located in NP
Qowowuyag (8188m) - located in CN
Dhaulagiri (8167m) - located in NP
Manaslu (8163m) - located in NP
Nanga Parbat (8125m) - located in PK
Annapurna I (8091m) - located in NP
Annapurna Himal (8091m) - located in NP
Gasherbrum Shan (8080m) - located in PK
Broad Feng (8051m) - located in PK
Gasherbrum II Feng (8034m) - located in PK
Xixabangma Feng (8027m) - located in CN
Broad Peak Central (8011m) - located in PK
*/

 /* Comment: This program may fail for large files, especially on 32-bit systems
  * where you have, at most, 3 GB to work with (of course, you have much less
  * in reality, and you need to map the file). However, the example file is small
  * enough that it should work on any system with 2GB RAM.
  */

/* The rest is up to you. Have fun. */
