/*  Chapter 14. ServerCP. SKELETON VERSION WITH EVEN MORE REMOVED
 *	Multi-threaded command line server. Named pipe version, COMPLETION PORT example
 *	Usage:	Server[UserName GroupName]
 *	ONE PIPE INSTANCE FOR EVERY CLIENT. 
 *  Assume MAX_SERVER_TH <= 64 (because of WaitForMultipleObject) */

#include "Everything.h"
#include "ClientServer.h" /* Request and response messages defined here */

typedef struct {				/* Argument to a server thread. */
	HANDLE hCompPort;			/* Completion port handle. */
	DWORD threadNum;
} SERVER_THREAD_ARG;
typedef SERVER_THREAD_ARG *LPSERVER_THREAD_ARG;

enum CP_CLIENT_PIPE_STATE { connected, requestRead, computed, responding, respondLast };
/*	Argument structure for each named pipe instance */
typedef struct {	/* Completion port keys refer to these structures */
} CP_KEY;

OVERLAPPED overLap;
volatile static int shutDown = 0;
static DWORD WINAPI Server (LPSERVER_THREAD_ARG);
static DWORD WINAPI ServerBroadcast (LPLONG);
static BOOL  WINAPI Handler (DWORD);
static DWORD WINAPI ComputeThread (PVOID);

static CP_KEY Key[MAX_CLIENTS_CP];

_tmain (int argc, LPTSTR argv[])
{
	HANDLE hCompPort, hMonitor, hSrvrThread[MAX_CLIENTS];
	DWORD iNp, iTh;
	DWORD AceMasks[] =	/* Named pipe access rights */
		{STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0X1FF, 0, 0 };
	SECURITY_ATTRIBUTES tempFileSA = {sizeof (SECURITY_ATTRIBUTES), NULL, TRUE};
	SERVER_THREAD_ARG ThArgs[MAX_SERVER_TH]; /* MAX_SERVER_TH <= 64 */
	OVERLAPPED ov = {0};

	if (!WindowsVersionOK (6, 0)) 
		ReportError (_T("This program requires Windows NT 6.0 or greater"), 1, FALSE);

	/* Console control handler to permit server shutDown */
	if (!SetConsoleCtrlHandler (Handler, TRUE))
		ReportError (_T("Cannot create Ctrl handler"), 1, TRUE);

	/* Pipe security is commented out for simplicity */
//	if (argc == 4)		/* Optional pipe security. */
//		pNPSA = InitializeUnixSA (0440, argv[1], argv[2], AceMasks, &hSecHeap);
			
	/* Create a thread broadcast pipe name periodically. */

	/*	Create an overlapped named pipe for every potential client, */
	/*	add to the completion port */
	/*	The assumption is that the maximum number of clients far exceeds */
	/*	the number of server threads	*/
	for (iNp = 0; iNp < MAX_CLIENTS_CP; iNp++) {
	}

	/*	Create server worker threads and a temp file name for each.*/
	for (iTh = 0; iTh < MAX_SERVER_TH; iTh++) {
	}
	_tprintf (_T("All server threads running.\n"));

	/* Wait for all server threads (<= 64) to terminate. */

	_tprintf (_T ("Server process will exit.\n"));
	return 0;
}

static DWORD WINAPI Server (LPSERVER_THREAD_ARG pThArg)
/* Server thread function. . */
{
	/* Each thread keeps its own request, response,
	and bookkeeping data structures on the stack. */

	DWORD nXfer;
	BOOL disconnect = FALSE;
	CP_KEY *pKey;
	RESPONSE response;
	OVERLAPPED serverOv = {0}, * pOv = NULL;

	/*	Main server thread loop to process completions */
	while (!disconnect && !shutDown) {	/* Wait for an outstanding operation to complete */
		
		/* Contrary to the documentation, this can return a false, indicating more data */
		if (shutDown) continue;

		/* The npState indicates which NP operation completed: connected, requestRead, ... */
		switch (pKey->npState) {
			case connected:
				{ /* A connection has completed, read a request */
					/* Open the temporary results file for this connection. */
					// _tprintf (_T("case connected.\n")); /* You may want to enable these trace statemetns */
					
				} 
			case requestRead:
				{ /* A read has completed. process the request */
					/* Create a separate thread to carry out the request asynchronously. */
					/* This server is then free to process other requests. */
				}
			case computed:
				{
					/* Results are in the temp file */
					/* Respond a line at a time. It is convenient to use
					C library line-oriented routines at this point. */
					// _tprintf (_T("case computed.\n"));
				}
			case responding:
				{
					// _tprintf (_T("case responding.\n"));
					/* In the process of responding a record at a time */
					/* Continue in this state until there are no more records */
				} 
			case respondLast:
				{
					// _tprintf (_T("case respondLast.\n"));
					/* Stay connected. */
					continue;
				} 
			default:
				{
					// _tprintf (_T("case default.\n"));
					/* Fatal for this thread. No recovery attempted (a good exercise!) */
					ReportError (_T("Server thread is in unknown state. Terminate this thread."), 0, FALSE);
					return 1;
				}
		} 

		FlushFileBuffers (pKey->hNp);
	}
	return 0;
}

static DWORD WINAPI ComputeThread (PVOID pArg)
{
	PROCESS_INFORMATION procInfo;
	STARTUPINFO startInfo;
	CP_KEY *pKey = (CP_KEY *)pArg;
	SECURITY_ATTRIBUTES tempSA = {sizeof(SECURITY_ATTRIBUTES), NULL, TRUE};
	GetStartupInfo (&startInfo);

	/* Open temp file, erasing previous contents */

	return 0;
}

static DWORD WINAPI ServerBroadcast (LPLONG pNull)
{
	MS_MESSAGE MsNotify;
	DWORD nXfer, iNp;
	HANDLE hMsFile;

	/* Open the mailslot for the MS "client" writer. */
	while (!shutDown) { /* Run as long as there are server threads */
		/* Wait for another client to open a mailslot. */
		Sleep (CS_TIMEOUT);
		hMsFile = CreateFile (MS_CLTNAME, GENERIC_WRITE | GENERIC_READ,
				FILE_SHARE_READ | FILE_SHARE_WRITE,
				NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hMsFile == INVALID_HANDLE_VALUE) continue;

		/* Send out the message to the mailslot. */

		MsNotify.msStatus = 0;
		MsNotify.msUtilization = 0;
		_tcscpy (MsNotify.msName, SERVER_PIPE);
		if (!WriteFile (hMsFile, &MsNotify, MSM_SIZE, &nXfer, NULL))
			ReportError (_T ("Server MS Write error."), 13, TRUE);
		CloseHandle (hMsFile);
	}

	_tprintf (_T("Shut down flag set. Cancel all outstanding I/O operations. NT6 dependency!\n"));
    /* Cancel all outstanding I/O. This is an NT6 dependency. What happens without this */
	for (iNp = 0; iNp < MAX_CLIENTS_CP; iNp++) {
		CancelIoEx (Key[iNp].hNp, NULL);
	}
	_tprintf (_T ("Shuting down monitor thread.\n"));

	_endthreadex (0);
	return 0;
}


BOOL WINAPI Handler (DWORD CtrlEvent)
{
	/* shutDown the system */
	_tprintf (_T("In console control handler\n"));
	InterlockedIncrement(&shutDown);
	return TRUE;
}
